@@ -1,4 +1,4 @@
-/* $Id: fetch.c,v 1.31 1996/06/23 21:33:52 agulbra Exp $ 
+/* $Id: fetch.c,v 1.37 1997/07/20 00:33:38 agulbra Exp $ 
 
 Written by Arnt Gulbrandsen <agulbra@troll.no> and copyright 1995
 Troll Tech AS, Postboks 6133 Etterstad, 0602 Oslo, Norway, fax +47
@@ -14,7 +14,6 @@ warranty, or liability of any kind.  */
 #include <ctype.h>
 #include <dirent.h>
 #include <fcntl.h>
-#include <malloc.h>
 #include <netdb.h>
 #include <netinet/in.h>
 #include <pwd.h>
@@ -54,6 +53,7 @@ void store( const char * filename,
 	    const char * references,
 	    const char * lines);
 
+int age( const char * date );
 int getgroup ( struct newsgroup *, int );
 int postarticles( void );
 void fixxover( void );
@@ -102,7 +102,7 @@ void  store( const char * filename,
 		    sprintf(tmp, "%d", ++cg->last);
 		    errno = 0;
 		    if ( verbose > 2 )
-			printf( " ...as article %d in %s\n",
+			printf( "..as article %d in %s\n",
 			       cg->last, cg->name );
 		} while ( link( filename, tmp )<0 && errno==EEXIST );
 		if ( errno )
@@ -112,6 +112,9 @@ void  store( const char * filename,
 		    sprintf( x, " %s:%d", cg->name, cg->last );
 		    x += strlen( x );
 		}
+	    } else {
+		if ( verbose )
+		    printf( ".. discarding unknown group %s\n", p );
 	    }
 	}
 	p = q;
@@ -119,7 +122,76 @@ void  store( const char * filename,
     if ( n > 1 )
 	fprintf( filehandle, "Xref: %s%s\n", fqdn, xrefincase );
 }
-		       
+
+
+
+int age( const char * date ) {
+    char monthname[4];
+    int month;
+    int year;
+    int day;
+    const char * d;
+    time_t tmp;
+    struct tm gmt;
+
+    if ( !date )
+	return 1000; /* large number: OLD */
+    d = date;
+    d += 5; /* "Date:" */
+    while( isspace( *d ) )
+	d++;
+
+    if ( isalpha( d[0] ) && isalpha( d[0] ) && isalpha( d[2] ) && 
+	 d[3] == ',' && isspace( d[4] ) )
+	d += 5; /* skip Mon, */
+
+    monthname[0] = '\0';
+
+    if ( sscanf( d, "%d %s %d", &day, monthname, &year ) < 3 )
+	return 1003;
+    else if ( ! ((year > 95 && year < 100) || (year > 1995)) )
+	return 1004;
+    else if ( strlen( monthname ) != 3 )
+	return 1005;
+    else if ( !(day > 0 && day < 32) )
+	return 1006;
+    else {
+	if ( !strcasecmp( monthname, "jan" ) )
+	    month = 0;
+	else if ( !strcasecmp( monthname, "feb" ) )
+	    month = 1;
+	else if ( !strcasecmp( monthname, "mar" ) )
+	    month = 2;
+	else if ( !strcasecmp( monthname, "apr" ) )
+	    month = 3;
+	else if ( !strcasecmp( monthname, "may" ) )
+	    month = 4;
+	else if ( !strcasecmp( monthname, "jun" ) )
+	    month = 5;
+	else if ( !strcasecmp( monthname, "jul" ) )
+	    month = 6;
+	else if ( !strcasecmp( monthname, "aug" ) )
+	    month = 7;
+	else if ( !strcasecmp( monthname, "sep" ) )
+	    month = 8;
+	else if ( !strcasecmp( monthname, "oct" ) )
+	    month = 9;
+	else if ( !strcasecmp( monthname, "nov" ) )
+	    month = 10;
+	else if ( !strcasecmp( monthname, "dec" ) )
+	    month = 11;
+	else {
+	    return 1001;
+	}
+	if ( year > 100 )
+	    year -= 1900; /* *sigh */
+	tmp = time(0);
+	gmt   = *(gmtime(&tmp));
+	return ((gmt.tm_year - year)*365 +
+		(gmt.tm_mon - month)*28 +
+		(gmt.tm_mday - day));
+    }
+}		       
 
 
 
@@ -148,8 +220,8 @@ int getgroup ( struct newsgroup * g, int server ) {
     reals = fakes = 0;
     chdirgroup( g->name );
 
-    fprintf( nntpout, "GROUP %s\r\n", g->name );
-    fflush( nntpout );
+    sprintf( lineout, "GROUP %s\r\n", g->name );
+    putaline();
 
     l = getaline( nntpin );
     if ( !l )
@@ -171,18 +243,35 @@ int getgroup ( struct newsgroup * g, int server ) {
     }
 
     if ( server > last+1 ) {
-	if ( verbose )
-	    printf( "switched upstream servers? %d > %d\n",
-		   server-1, last );
-	server = window; /* insane - recover thoroughly */
+	syslog( LOG_INFO, "last seen article was %d, server now has %d-%d",
+		server, window, last );
+	if ( server > last+5 ) {
+	    if ( verbose )
+		printf( "switched upstream servers? %d > %d\n",
+			server-1, last );
+	    server = window; /* insane - recover thoroughly */
+	} else {
+	    if ( verbose )
+		printf( "rampant spam cancel? %d > %d\n",
+			server-1, last );
+	    server = last - 5; /* a little bit too much */
+	}
     }
     if ( artlimit && last-server > artlimit ) {
 	if ( verbose )
-	    printf( "skipping articles %d-%d inclusive\n", 
+	    printf( "skipping articles %d-%d inclusive (article limit)\n", 
 		    server, last-artlimit-1 );
 	server = last-artlimit;
     }
 
+    if ( initiallimit && (g->first >= g->last) && 
+	 last-server > initiallimit ) {
+	if ( verbose )
+	    printf( "skipping articles %d-%d inclusive (initial limit)\n", 
+		    server, last-initiallimit-1 );
+	server = last-initiallimit;
+    }
+
     sprintf( s, "%s/interesting.groups/%s", spooldir, g->name );
     if ( stat( s, &st ) )
 	return last; /* race condition: I hope this is proper */
@@ -194,7 +283,7 @@ int getgroup ( struct newsgroup * g, int server ) {
 	    printf("skipping %s from now on\n", g->name);
 	syslog( LOG_INFO, "skip %s: %d, %d", g->name, server, last );
 	unlink( s );
-    }
+   }
 
     if ( window < server )
 	window = server;
@@ -206,9 +295,9 @@ int getgroup ( struct newsgroup * g, int server ) {
 	if ( verbose )
 	    printf( "%s: considering articles %d-%d\n", 
 		    g->name, server, last );
-	fprintf( nntpout, "XHDR Message-ID %d-%d\r\n",
+	sprintf( lineout, "XHDR Message-ID %d-%d\r\n",
 		 server, last );
-	fflush( nntpout );
+        putaline();
 	if ( nntpreply() == 221 ) {
 	    stufftoget = (char*) realloc( stufftoget, last + 1 - server );
 	    if ( stufftoget )
@@ -245,18 +334,19 @@ int getgroup ( struct newsgroup * g, int server ) {
 		    window <= last &&
 		    !stufftoget[ window - server ] ) {
 		if ( verbose > 3 )
-		    printf( "%s: skipping %d\n", g->name, window );
+		    printf( "%s: skipping %d - not available or too old\n",
+			    g->name, window );
 		window++;
 	    }
 	    if ( window <= last ) {
 		outstanding++;
 		if ( verbose > 1 ) 
-		    printf( "%s: ARTICLE %d (%d outstanding)\n",
+		    printf( "%s: ARTICLE %d (%d up in the air)\n",
 			    g->name, window, outstanding );
-		fprintf(nntpout, "ARTICLE %d\r\n", window++);
+		sprintf(lineout, "ARTICLE %d\r\n", window++);
+		putaline();
 	    }
 	}
-	fflush(nntpout);
 
 	if ( outstanding > 0 ) {
 	    l = getaline( nntpin );
@@ -266,13 +356,13 @@ int getgroup ( struct newsgroup * g, int server ) {
 	    outstanding--;
 	    if (sscanf(l, "%3d %d", &n, &h) < 2 || n != 220 ) {
 		if ( verbose )
-		    printf( "%s: reply %03d (%d outstanding)\n",
+		    printf( "%s: reply %03d (%d more up in the air)\n",
 			    g->name, n, outstanding );
 		continue;
 	    }
 
 	    if ( verbose )
-		printf( "%s: receiving article %d (%d outstanding)\n",
+		printf( "%s: receiving article %d (%d more up in the air)\n",
 			g->name, h, outstanding );
 
 	    for ( h=0; h<10; h++ ) {
@@ -288,11 +378,20 @@ int getgroup ( struct newsgroup * g, int server ) {
 		    for ( h=0; hd[h] && *(hd[h]) && h<4; h++ )
 			;
 		    f = NULL;
-		    if ( h == 4 ) {
+		    if ( h != 4 ) {
+			if ( verbose )
+			    printf( "discarding it - no message-id found\n" );
+		    } else if ( age( hd[5] ) > 10 ) {
+			if ( verbose )
+			    printf( "%s posted more than %d days ago: %s",
+				    hd[1], 10, hd[5] );
+		    } else {
 			c = lookup(strchr( hd[1], '<'));
 			if ( stat( c, &st ) && errno == ENOENT )
 			    f = fopen( c, "w" );
-			/* ie. discard cross-posted stuff and other garbage */
+			else if ( verbose )
+			    printf( "discarding it - already have it or "
+				    "file system error\n" );
 		    }
 		    if (f) {
 			for ( h=0; h<10; h++ )
@@ -329,6 +428,8 @@ int getgroup ( struct newsgroup * g, int server ) {
 			       *hd[7] ? hd[7]+strlen(hnames[7]) : "");
 		    }
 		    do {
+			if ( l && *l == '.' )
+			    l++;
 			if (f)
 			    fprintf(f, "%s\n", l);
 		    } while ( ((l=getaline(nntpin)) != NULL) &&
@@ -373,6 +474,8 @@ int getgroup ( struct newsgroup * g, int server ) {
 			strcat( hd[n], l );
 			strcat( hd[n], "\n" );
 			h = n;
+			if ( verbose > 4 && hnames[n] && *hnames[n] )
+			    printf( "...saw header %s\n", hnames[n] );
 		    }
 		}
 	    }
@@ -382,23 +485,27 @@ int getgroup ( struct newsgroup * g, int server ) {
     return last +1;
 }
 
-
-
+/*
+ * get active file from remote server
+ */
 static void nntpactive( void ) {
     char * l;
     int first;
     struct stat st;
 
     sprintf( s, "%s/active.read", spooldir );
-    if ( stat( s, &st )==0 && now-st.st_mtime < TIMEOUT_LONG ) {
+    if ( active && stat( s, &st )==0 && now-st.st_mtime < TIMEOUT_LONG ) {
 	if ( verbose )
 	    printf( "LIST ACTIVE done only %d seconds ago, skipping\n",
 		    (int)(now-st.st_mtime) );
 	return;
     }
 
-    fprintf(nntpout, "LIST ACTIVE\r\n");
-    fflush(nntpout);
+    if ( verbose )
+	printf( "Reading list of active newsgroups\n" );
+
+    sprintf(lineout, "LIST ACTIVE\r\n");
+    putaline();
 
     if ( nntpreply() != 215 )
 	return;
@@ -419,6 +526,37 @@ static void nntpactive( void ) {
 	}
     }
 
+    if ( verbose )
+	printf( "Reading newsgroup descriptions\n" );
+
+    /*
+     * add descriptions
+     */
+    fprintf(nntpout, "LIST NEWSGROUPS\r\n");
+    fflush(nntpout);
+    if ( nntpreply() == 215 ) {
+	while ( (l=getaline(nntpin)) && ( *l != '.' ) ) {
+	    char * p;
+	    struct newsgroup * g;
+
+	    p = l;
+	    while ( !isspace( *p ) )
+		p++;
+	    if (*p)
+		*p++ = '\0';
+	    while( isspace( *p ) )
+		p++;
+	    if ( (g = findgroup(l)) != NULL &&
+		 strcmp(p, "?" ) && strcmp(p, "-x-") ) {
+		if ( verbose )
+		    printf( "new description of %s: %s\n", g->name, p );
+		g->desc = strdup(p);
+	    } else if ( verbose ) {
+		printf( "didn't work out: %s: %s\n", l, p );
+	    }
+	}
+    }
+
     sprintf( s, "%s/active.read", spooldir );
     unlink( s );
     close( open( s, O_WRONLY|O_CREAT, 0664 ) );
@@ -461,17 +599,22 @@ int postarticles( void ) {
 	if ( !stat(de->d_name, &st) &&
 	     S_ISREG( st.st_mode ) &&
 	     (f=fopen( de->d_name, "r" )) != NULL ) {
-	    fprintf( nntpout, "POST\r\n" );
-	    fflush( nntpout );
+	    if ( verbose )
+		printf( "Posting %s...\n", de->d_name );
+	    sprintf( lineout, "POST\r\n" );
+	    putaline();
 	    r = nntpreply();
 	    if ( r == 340 ) {
 		do {
 		    line = getaline( f );
-		    if ( line )
-			fprintf( nntpout, "%s\r\n", line );
+		    if ( line ) {
+			sprintf( lineout, "%s\r\n", line );
+			putaline();
+		    }
 		} while ( line && strcmp( line, "." ) );
-		fflush( nntpout );
 		if ( !line ) {
+		    if ( verbose )
+			printf( " - failed: %03d reply to POST\n", r );
 		    sprintf( s, "%s/failed.postings", spooldir );
 		    mkdir( s, 0775 );
 		    sprintf( s, "%s/failed.postings/%s", 
@@ -489,12 +632,18 @@ int postarticles( void ) {
 		}
 		line = getaline( nntpin );
 		if ( line && !strncmp( line, "240", 3 ) ) {
+		    if ( verbose )
+			printf( " - OK\n" );
 		    syslog( LOG_INFO, "posted article" ); /* useless */
 		    unlink( de->d_name);
 		} else {
 		    if ( line && !strncmp( line, "441 435", 7 ) ) {
+			if ( verbose )
+			    printf(" - upstream server had that message-id\n");
 			unlink( de->d_name );
 		    } else {
+			if ( verbose )
+			    printf( " - article rejected: %s\n", line );
 			sprintf( s, "%s/failed.postings", spooldir );
 			mkdir( s, 0775 );
 			sprintf( s, "%s/failed.postings/%s", 
@@ -567,19 +716,20 @@ void ihave( void ) {
 	    } while ( line && strncasecmp( line, "Message-ID: ", 12 ) );
 	    if ( line && strlen( line+12 ) ) {
 		char * msgid;
-		/* bad assumption here - IHAVE\n\t<asdf@asdf> is legal
-		    according to the RFCs */
-		msgid = strdup( line+12 );    
-		fprintf( nntpout, "IHAVE %s\r\n", msgid );
-		fflush( nntpout );
+		msgid = line+12;
+		while( msgid && *msgid && isspace(*msgid) )
+		    msgid++;
+		msgid = strdup( msgid );
+		sprintf( lineout, "IHAVE %s\r\n", msgid );
+		putaline();
 		if ( nntpreply() == 335 ) {
 		    fseek( f, 0, SEEK_SET );
 		    do {
 			line = getaline( f );
 			if ( line )
-			    fprintf( nntpout, "%s\r\n", line );
+			    sprintf( lineout, "%s\r\n", line );
+                            putaline();
 		    } while ( line && strcmp( line, "." ) );
-		    fflush( nntpout );
 		    syslog( LOG_INFO, "IHAVE %s received %03d reply",
 			    msgid, nntpreply() );
 		}
@@ -611,17 +761,20 @@ static void processactive( void ) {
 	    if ( g ) {
 		g->alive = 1;
 		g->server = getgroup( g, g->server );
+		syslog( LOG_INFO, "%s: fetched to %d", g->name, g->server );
 	    }
 	}
     }
     rewinddir( d );
-    while ( (de=readdir(d)) ) {
-	if ( '.' == *(de->d_name) && isalpha( (de->d_name)[1] ) ) {
-	    /* someone did GROUP but didn't read anything.. feed them a fake */
-	    g = findgroup( 1+(de->d_name) );
+    while ( (de=readdir(d)) ) {    
+	if ( de->d_name[0] == '.' && isalpha( de->d_name[1] ) ) {
+	    g = findgroup( de->d_name + 1 );
+	    if ( verbose > 3 && (!g || !g->alive) )
+		printf( "considering %s - %s upstream\n", de->d_name+1,
+			g ? " exists" : "does not exist" );
 	    if ( g && !g->alive ) {
+		g->alive = 1;
 		g->last++;
-		unlink( de->d_name );
 	    }
 	}
     }
@@ -631,47 +784,82 @@ static void processactive( void ) {
 
 
 static void processsupplement( const char * supplement ) {
-    char newfile[20];
-    FILE * i;
-    FILE * o;
-    struct newsgroup * root;
+    FILE * f;
+    DIR * d;
+    char * l;
+    struct dirent * de;
+    struct newsgroup * g;
+    struct stat st;
+    int fd;
+    int server = 0, havefile = 0;
 
-    root = NULL;
+    static char * stuff ;
+
+    /* read supplementary server info */
     sprintf( s, "%s/%s", libdir, supplement );
-    i = fopen( s, "r" );
-    if ( i ) {
-	DIR * d;
-	struct dirent * de;
-	struct newsgroup * g;
-
-	strcpy( newfile, ".supplement.XXXXX" );
-	o = fopen( mktemp( newfile ), "w" );
-	
-	sprintf( s, "%s/interesting.groups", 
-		 spooldir );
-	d = opendir( s );
-	if ( !d ) {
-	    syslog( LOG_ERR, "opendir %s: %m", s );
-	    return;
+    fd = open( s, O_RDONLY );
+    if ( stat( s, &st ) == 0 ) {
+        if ( verbose > 3 )
+            printf( "Read supplementary server info from %s\n", s );
+        stuff = critmalloc( st.st_size+1, "Reading supplementary server info" );
+        if ( (fd=open( s, O_RDONLY))<0 ||
+             (read( fd, stuff, st.st_size ) < st.st_size) ) {
+            syslog( LOG_ERR, "can't open/read %s: %m", s );
+	    havefile = 0;
+            free( stuff );
+        } else {
+	    havefile = 1;
+            close( fd );
+            stuff[st.st_size] = '\0'; /* 0-terminate string */
 	}
-
-	while ( (de=readdir(d)) ) {
-	    if ( isalpha( *(de->d_name) ) ) {
-		g = findgroup( de->d_name );
-		if ( g ) {
-		    g->alive = 1;
-		}
-	    }
-	}
-	closedir( d );
     }
+
+    sprintf( s, "%s/interesting.groups", 
+	     spooldir );
+    d = opendir( s );
+    if ( !d ) {
+	syslog( LOG_ERR, "opendir %s: %m", s );
+	return;
+    }
+
+    sprintf( s, "%s/%s", libdir, supplement );
+    f = fopen( s, "w" );
+    if ( ( f == NULL ) && verbose )
+        printf( "Couldn't open %s for writing.\n", s );
+    while ( (de=readdir(d)) ) {
+	if ( isalpha( *(de->d_name) ) ) {
+	    g = findgroup( de->d_name );
+	    if ( g ) {
+	        g->alive = 1;
+		if ( havefile == 1 ) {
+		/* TODO: this doesn't work properly */
+		    l = strstr( stuff, g->name );
+		    if ( l == NULL )
+		        server = 0;
+		    else {
+		        l = strchr( l, ' ' );
+		        server = strtol( l, NULL, 10 );
+		    }
+		}
+		else
+		    server = 0;
+#ifdef BROKEN
+		server = getbymsgid(g, server);
+#endif
+		if ( f != NULL )
+                    fprintf( f, "%s %d\n", de->d_name, server );
+            }
+	}
+    }
+    closedir( d );
+    if ( f != NULL )
+        fclose(f);
+
+    free( stuff );
 }
 
 
 
-
-
-
 void fixxover( void ) {
     DIR * d;
     struct dirent * de;
@@ -692,14 +880,17 @@ void fixxover( void ) {
     }
     closedir( d );
 }
-    
+
 
 
 int main( int argc, char ** argv ) {
     struct passwd * pw;
     int option;
+    FILE *lf;
+    int usesupplement;
 
     verbose = 0;
+    usesupplement = 0;
 
     pw = getpwnam( "news" );
     if ( !pw ) {
@@ -707,7 +898,7 @@ int main( int argc, char ** argv ) {
 	exit( 1 );
     }
 
-    setgid( pw->pw_gid );
+    setregid( pw->pw_gid, pw->pw_gid );
     setreuid( pw->pw_uid, pw->pw_uid );
 
     if ( getuid() != pw->pw_uid || getgid() != pw->pw_gid ) {
@@ -715,7 +906,12 @@ int main( int argc, char ** argv ) {
 	exit( 1 );
     }
 
-    while ( (option=getopt( argc, argv, "vx:" )) != -1 ) {
+    if ( (lf = fopen(lockfile, "r+")) != NULL ) {
+        fprintf( stderr, "lockfile %s exists, abort fetch ...\n", lockfile);
+        exit( 1 );
+    }
+
+    while ( (option=getopt( argc, argv, "lvx:" )) != -1 ) {
 	if ( option == 'v' ) {
 	    verbose++;
 	} else if ( option == 'x' ) {
@@ -724,16 +920,27 @@ int main( int argc, char ** argv ) {
 	    endptr = NULL;
 	    extraarticles = strtol( nptr, &endptr, 0 );
 	    if ( !nptr || !*nptr || !endptr || *endptr || !extraarticles ) {
-		fprintf( stderr, "Usage: fetch [-v] [-x #]\n"
+		fprintf( stderr, "Usage: fetch [-v] [-x #] [-l]\n"
 			 "  -v: more verbose (may be repeated)\n"
-			 "  -x: check for # extra articles in each group\n" );
+			 "  -x: check for # extra articles in each group\n"
+			 "  -l: don't use supplementary servers\n" );
+                unlink( lockfile );
 		exit( 1 );
 	    }
+	} else if ( option == 'l' ) {
+	    usesupplement = 0;
 	}
     }
 
-    if ( verbose )
-	printf( "verbosity level is %d\n", verbose );
+    if ( verbose ) {
+	printf( "%s: verbosity level is %d\n", version, verbose );
+	if ( verbose > 3 ) {
+	    printf("Don't fetch threads that nobody has read after %d days\n",
+	           TIMEOUT_LONG/3600/24);
+	    printf("Don't fetch threads that have been read once after %d days\n",
+	           TIMEOUT_SHORT/3600/24);
+        }
+    }
 
     whoami();
 
@@ -745,50 +952,63 @@ int main( int argc, char ** argv ) {
 
     readconfig();
     if ( upstream == NULL ) {
+	if ( verbose )
+	    printf( "can't find an upstream server\n" );
 	syslog( LOG_ERR, 
 		"no server name in %s", s );
+        unlink( lockfile );
 	exit( 2 );
     }
 
-    do {
-	if (!nntpconnect()) {
-	    syslog( LOG_ERR, "unable to connect to %s", upstream );
-	    fprintf( stderr, "unable to connect to %s\n", upstream );
-	    exit(1);
-	}
-	ihave();
-	fprintf( nntpout, "MODE READER\r\n" );
-	fflush( nntpout );
-	if ( nntpreply() == 498 )
-	    continue;
-    } while ( !postarticles() );
+    if (!nntpconnect()) {
+        syslog( LOG_ERR, "unable to connect to %s", upstream );
+        fprintf( stderr, "unable to connect to %s\n", upstream );
+    } else {
+        do {
+	    ihave();
+	    sprintf( lineout, "MODE READER\r\n" );
+	    putaline();
+	    fflush( nntpout );
+	    if ( nntpreply() == 498 )
+	        continue;
+        } while ( !postarticles() );
+	if (verbose)
+	    printf("Connecting to %s ...\n", upstream);
+	readactive();	/* read groupinfo file */
+	nntpactive();	/* get information from remote server */
+	processactive();
 
-    readactive();
-    nntpactive();
-    processactive();
+	sprintf( lineout, "QUIT\r\n"); /* say it, then just exit :) */
+	putaline();
+	writeactive();	/* write groupinfo file */
+    }
 
-    fprintf(nntpout, "QUIT\r\n"); /* say it, then just exit :) */
-    fflush(nntpout);
-    writeactive();
-
-    while ( supplement ) {
+    while ( usesupplement && supplement ) {
 	if ( nntpout )
 	    fclose( nntpout );
 	if ( nntpin )
 	    fclose( nntpin );
 	nntpin = nntpout = NULL;
 	upstream = supplement->name;
+	if ( verbose )
+	    printf("Trying to connect to %s ... ", supplement->name);
 	if ( nntpconnect() ) {
-	    syslog( LOG_INFO, "supplementing from %s", supplement->name );
+	    if (verbose)
+	        printf("connected.\n");
 	    ihave();
-	    fprintf( nntpout, "MODE READER\r\n" );
+	    sprintf( lineout, "MODE READER\r\n" );
 	    fflush( nntpout );
 	    if ( nntpreply() != 498 ) {
 		(void) postarticles();
 		processsupplement( supplement->name );
 	    }
-	    fprintf(nntpout, "QUIT\r\n"); /* say it, then just exit :) */
-	    fflush(nntpout);
+	    sprintf( lineout, "QUIT\r\n"); /* say it, then just exit :) */
+	    putaline();
+	    writeactive();
+	}
+	else {
+	    if ( verbose )
+	        printf("failed.\n");
 	}
 	supplement = supplement->next;
     }
@@ -797,10 +1017,11 @@ int main( int argc, char ** argv ) {
 
     if ( verbose || (fork() <= 0) ) {
 #if defined(PRIO_MAX)
-	setpriority( PRIO_PROCESS, 0, PRIO_MAX/2 );
+        setpriority( PRIO_PROCESS, 0, PRIO_MAX/2 );
 #endif
-	fixxover();
+        fixxover();
     }
 
+    unlink( lockfile );
     exit(0);
 }
