@@ -1,5 +1,4 @@
-static char rcsid[] = 
-"$Id: nntpd.c,v 1.19 1996/06/23 21:33:53 agulbra Exp $";
+/* $Id: nntpd.c,v 1.27 1997/07/21 14:08:40 agulbra Exp $ */
 
 /* Written by Arnt Gulbrandsen <agulbra@troll.no> and copyright 1995
 Troll Tech AS, Postboks 6133 Etterstad, 0602 Oslo, Norway, fax +47
@@ -8,8 +7,8 @@ Troll Tech AS, Postboks 6133 Etterstad, 0602 Oslo, Norway, fax +47
 Use, modification and distribution is allowed without limitation,
 warranty, or liability of any kind.  */
 
-#include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <ctype.h>
@@ -33,13 +32,14 @@ warranty, or liability of any kind.  */
 int write2(int fd, const char *msg);
 int hash (const char *);
 FILE * fopenart(const char *);
+FILE * fopenpseudoart(const char * arg, const int article_num);
 void list(struct newsgroup *, int);
 void rereadactive(void);
 
-const char* rfctime(void);
 void parser(void);
 void error(const char *msg);
 void doarticle(const char *, int);
+void dostat(const char *);
 void dogroup(const char *);
 void dohelp(void);
 void domove(int);
@@ -60,51 +60,9 @@ int artno = 0; /* current article number, initially 0 and almost unused */
 char *cmd; /* current command line */
 struct newsgroup * active;
 time_t activetime;
-char version[10];
 char fqdn[64];
 static int justaftergroup;
 
-const char* rfctime(void) {
-    static char date[128];
-    const char * months[] = { "Jan", "Feb", "Mar", "Apr",
-	"May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
-    const char * days[] = { "Sun", "Mon", "Tue", "Wed",
-	"Thu", "Fri", "Sat" };
-    time_t now;
-    struct tm gmt;
-    struct tm local;
-    struct tm conv;
-    signed int diff;
-
-    /* get local and Greenwich times */
-    now = time(0);
-    gmt   = *(gmtime(&now));
-    local = *(localtime(&now));
-    conv  = *(localtime(&now));
-
-    /* if GMT were local, what would it be? */
-    conv.tm_year = gmt.tm_year;
-    conv.tm_mon  = gmt.tm_mon;
-    conv.tm_mday = gmt.tm_mday;
-    conv.tm_hour = gmt.tm_hour;
-    conv.tm_min  = gmt.tm_min;
-    conv.tm_sec  = gmt.tm_sec;
-    conv.tm_wday = gmt.tm_wday;
-    conv.tm_yday = gmt.tm_yday;
-
-    /* see how far the real local is from GMT-as-local */
-    /* should convert to signed int from double properly */
-    diff = (mktime(&local) - mktime(&conv)) / 60.0;
-
-    /* finally print the string */
-    sprintf(date, "%3s, %d %3s %4d %02d:%02d:%02d %+03d%02d",
-	    days[local.tm_wday], local.tm_mday, months[local.tm_mon],
-	    local.tm_year+1900, local.tm_hour, local.tm_min, local.tm_sec, 
-	    diff/60, diff%60);
-
-    return (date);
-}
-
 
 void rereadactive(void) {
     struct stat st;
@@ -114,6 +72,7 @@ void rereadactive(void) {
 
     if ( (!stat(s, &st) && (st.st_mtime > activetime)) ||
 	 (active==NULL) ) {
+	syslog( LOG_DEBUG, "rereading %s", s );
 	readactive();
 	activetime = st.st_mtime;
     }
@@ -143,18 +102,21 @@ void parser(void) {
 	    n++;
 	}
 
-	if (n>999 || n==0) {
+	if ( n == 0 )
+	    continue; /* outside the spec, but let's do it anyway */
+
+	if ( n>999 ) {
 	    printf("400 Dazed and confused\r\n");
 	    return;
 	}
 
-	syslog( LOG_DEBUG, "%s", cmd );
-
 	while(isspace(cmd[n]) && n<1000)
 	    cmd[n++] = '\0';
 
 	arg = cmd+n;
 
+	syslog( LOG_DEBUG, "%s %s", cmd, arg);
+
 	while(cmd[n] && n<1000)
 	    n++;
 	n--;
@@ -165,11 +127,11 @@ void parser(void) {
 	if (!strcasecmp(cmd, "article")) {
 	    doarticle(arg, 3);
 	} else if (!strcasecmp(cmd, "head")) {
-	    doarticle(arg, 1);
-	} else if (!strcasecmp(cmd, "body")) {
 	    doarticle(arg, 2);
+	} else if (!strcasecmp(cmd, "body")) {
+	    doarticle(arg, 1);
 	} else if (!strcasecmp(cmd, "stat")) {
-	    printf("500 STAT is neither sensible nor fully specified\r\n");
+	    doarticle(arg, 0);
 	} else if (!strcasecmp(cmd, "help")) {
 	    dohelp();
 	} else if (!strcasecmp(cmd, "ihave")) {
@@ -208,6 +170,32 @@ void parser(void) {
 }
 
 
+/* open a pseudo art */
+FILE * fopenpseudoart(const char * arg, const int article_num) {
+    FILE *f = NULL;
+    char msgidbuf[128];
+    char * c;
+    struct newsgroup * g;    
+
+    if (article_num && (article_num < group->first)) {
+	f = buildpseudoart(group->name);
+    } else if (!article_num) {
+	if (!strncmp(arg, "<leafnode:placeholder:", 22)) {
+	    strncpy(msgidbuf, arg+22, 128);
+	    msgidbuf[127] = '\0';
+	    if ((c = strchr(msgidbuf, '@')) != NULL) {
+		*c = '\0';
+		g = findgroup(msgidbuf);
+		if (g) {
+		    f = buildpseudoart(g->name);
+		}
+	    }
+	}
+    }
+    return f;
+}
+
+
 /* open an article by number or message-id */
 FILE * fopenart(const char * arg) {
     unsigned long int a;
@@ -216,17 +204,24 @@ FILE * fopenart(const char * arg) {
 
     t = NULL;
     a = strtol(arg, &t, 10);
-    if (t && !*t && group) {
+    if (a && t && !*t && group) {
 	f = fopen(arg, "r");
 	if (f)
 	    artno = a;
+	else
+	    f = fopenpseudoart(arg, a);
 	markinterest();
 	/* else try message-id lookup */
     } else if ( arg && *arg=='<' ) {
 	f = fopen(lookup(arg), "r");
+	if ( !f )
+	    f = fopenpseudoart(arg, a);
     } else if (group && artno) {
 	sprintf(s, "%d", artno);
+	printf( "fopenart: %s\n", s );
 	f = fopen(s, "r");
+	if (!f)
+	    f = fopenpseudoart(s, a);
 	markinterest();
     } else {
 	f = NULL;
@@ -242,6 +237,11 @@ void doarticle(const char *arg, int what) {
     char *p;
     long yuck;
 
+    if ( !group ) {
+	printf( "412 No newsgroup selected\r\n" );
+	return;
+    }
+
     f = fopenart(arg);
     if (!f) {
 	if ( strlen( arg ) )
@@ -256,14 +256,12 @@ void doarticle(const char *arg, int what) {
 	yuck = 0;
 
     printf("%3d %ld Text follows, then a line consisting of a single '.'\r\n",
-	   what==3 ? 220 : 220+what, yuck ); /* nice idea but... */
+	   223-what, yuck );
 
     while (fgets(s, 1024, f) && *s && (*s!='\n')) {
-	if (what & 1) {
+	if (what & 2) {
 	    p = s;
-	    while(p && *p)
-		p++;
-	    if (*--p == '\n')
+	    if ( (p = strchr(p, '\n')) )
 		*p = '\0';
 
 	    printf("%s%s\r\n", *s=='.' ? "." : "", s); /* . on headers :( */
@@ -273,21 +271,21 @@ void doarticle(const char *arg, int what) {
     if (what==3)
 	printf("\r\n"); /* empty separator line */
 
-    while ((what & 2) && fgets(s, 1024, f) && *s) {
+    while ((what & 1) && fgets(s, 1024, f) && *s) {
 	p = s;
-	while(p && *p)
-	    p++;
-	if (*--p == '\n')
+	if ( (p = strchr(p, '\n')) )
 	    *p = '\0';
 
 	printf("%s%s\r\n", *s=='.' ? "." : "", s);
     }
-    printf(".\r\n");
+    if ( what )
+	printf(".\r\n");
     fclose(f);
 
     return; /* OF COURSE there were no errors */
 }
 
+
 /* note bug.. need not be _immediately after_ GROUP */
 void markinterest( void ) {
     int f;
@@ -305,26 +303,26 @@ void markinterest( void ) {
 /* change to group - note no checks on group name */
 void dogroup(const char *arg) {
     struct newsgroup * g;
-
-    if ( justaftergroup && group && group->name ) {
-	int f;
-	/* mark the previous group as _possibly_ interesting */
-	sprintf( s, "%s/interesting.groups/.%s", spooldir, group->name );
-	if ( (f=open( s, O_WRONLY|O_CREAT|O_TRUNC, 0644)) >= 0 )
-	    close(f);
-    }
+    int f;
 
     rereadactive();
     g = findgroup(arg);
     if (g) {
 	group = g;
 	chdirgroup( g->name );
-	printf("211 %d %d %d %s group selected\r\n", 
-	       g->last >= g->first ? g->last-g->first+1 : 0, 
-	       g->first, g->last, g->name);
+	if (g->last < g->first)
+	    printf("211 %d %d %d %s group selected (pseudo article)\r\n", 
+		   1, g->last, g->last, g->name);
+	else
+	    printf("211 %d %d %d %s group selected\r\n", 
+		   g->last >= g->first ? g->last-g->first+1 : 0, 
+		   g->first, g->last, g->name);
 	artno = g->first;
 	fflush( stdout );
 	justaftergroup = TRUE;
+	sprintf( s, "%s/interesting.groups/.%s", spooldir, group->name );
+	if ( (f=open( s, O_WRONLY|O_CREAT|O_TRUNC, 0644)) >= 0 )
+	    close(f);
     } else {
 	printf("411 No such group\r\n");
     }
@@ -359,10 +357,16 @@ void domove( int by ) {
 void list(struct newsgroup *g, int what) {
     if (g) {
 	list (g->right, what);
-	if (what)
-	    printf( "%s %s\r\n", g->name, g->desc ? g->desc : "" );
-	else
-	    printf( "%s %010d %010d y\r\n", g->name, g->last, g->first);
+	if (what) {
+	    printf( "%s\t%s\r\n", g->name, g->desc ? g->desc : "" );
+	} else {
+	    /* If the group is empty, but has server arts, fudge a
+               dummy article in... */
+	    if (g->last < g->first)
+		printf( "%s %010d %010d y\r\n", g->name, g->last, g->last);
+	    else
+		printf( "%s %010d %010d y\r\n", g->name, g->last, g->first);
+	}
 	list (g->left, what);
     }
 }
@@ -377,15 +381,22 @@ void dolist(const char *arg) {
 	printf("503 Group information file does not exist!\r\n");
     else {
 	if ( !*arg || !strcasecmp(arg, "active") ) {
-	    printf("215 Brace yourself\n");
+	    printf("215 Newsgroups in form \"group high low flags\".\r\n");
 	    list ( active, 0 );
 	    printf(".\r\n");
 	} else if ( !strcasecmp(arg, "newsgroups") ) {
-	    printf("215 Brace yourself\n");
+	    printf("215 Descriptions in form \"group description\".\r\n");
 	    list ( active, 1 );
 	    printf(".\r\n");
+	} else if ( !strcasecmp(arg, "extensions") ) {
+	    printf("501 No extensions available.\r\n" );
+	} else if ( !strcasecmp(arg, "active.times") ) {
+	    printf("215 Placeholder - Leafnode will fetch groups on demand\r\n"
+		   "news.announce.newusers 42 tale@uunet.uu.net\r\n"
+		   "news.answers 42 tale@uunet.uu.net\r\n"
+		   ".\r\n");
 	} else {
-	    printf( "502 Syntax error\r\n" );
+	    printf( "501 Syntax error\r\n" );
 	}
     }
 }
@@ -463,43 +474,6 @@ void dopost( void ) {
     sprintf( outname, "%s/out.going/%d-%d-%d", 
 	     spooldir, (int)getpid(), (int)time(NULL), ++postingno );
 
-    if ( 0 && upstream && ( nntpout || nntpconnect() ) ) {
-	int r;
-	fprintf( nntpout, "POST\r\n" );
-	fflush( nntpout );
-	r = nntpreply();
-	printf( "%3d reply received from %s\r\n", r, upstream );
-	fflush( stdout );
-	if ( r == 340 ) {
-	    do {
-		int inheaders = 1;
-
-		if ( (line = getaline( stdin )) )
-		    if ( inheaders && !havepath && 
-			 !strncmp( line, "Path: ", 6 ) ) {
-			fprintf( nntpout, "Path: %s!%s\r\n",
-				 fqdn, line[6] ? line+6 : "nobody" );
-			havepath = 1;
-		    } else if ( inheaders && !*line ) {
-			inheaders = 0;
-			if ( !havepath ) {
-			    fprintf( nntpout, "Path: %s!nobody\r\n\r\n", fqdn);
-			    havepath++; /* duh? why? */
-			} else {
-			    fprintf( nntpout, "\r\n" );
-			}
-		    } else {
-			fprintf( nntpout, "%s\r\n", line );
-		    }
-	    } while ( line && strcmp( line, "." ) );
-	    fflush( nntpout );
-	    line = getaline( nntpin );
-	    printf( "%s\r\n", line );
-	    syslog( LOG_INFO, "relayed posting, reply %s", line );
-	    return;
-	}
-    }
-
     out = open( outname, O_WRONLY|O_EXCL|O_CREAT, 0444 );
     if ( out < 0 ) {
 	char dir[80];
@@ -512,7 +486,7 @@ void dopost( void ) {
 	}
     }
 
-    printf( "340 Go ahead..\r\n" );
+    printf( "340 Go ahead.\r\n" );
     fflush( stdout );
 
     do {
@@ -581,8 +555,8 @@ void dopost( void ) {
 	    }
 	    if ( !havemessageid ) {
 		char tmp[80];
- 		sprintf( tmp, "Message-ID: %s\r\n",
- 			 generateMessageID() );
+		sprintf( tmp, "Message-ID: %s\r\n",
+			 generateMessageID() );
 		write( out, tmp, strlen( tmp ) );
 	    }
 	}
@@ -603,7 +577,9 @@ void dopost( void ) {
 	    line[--len] = '\0';
 	if ( len && line[len-1]=='\r' )
 	    line[--len] = '\0';
-	if ( len )
+	if ( len > 1 && line[0] == '.' )
+	    write( out, line+1, len-1 );
+	else if ( len )
 	    write( out, line, len );
 	write( out, "\r\n", 2 );
     } while ( line[0] != '.' || line[1] != '\0' );
@@ -621,15 +597,15 @@ void dopost( void ) {
 
 
 
-
-
 void doxhdr(char *arg) {
     char * h[] = { "Subject", "From", "Date", "Message-ID", "References",
-		   "Bytes", "Lines" };
+		   "Bytes", "Lines"
+    }; /* ugly emacs workaround */
+
     int n = 7;
     char * l;
     unsigned long int a,b,c;
-    char buffer[1024];
+    FILE * f;
 
     l = arg;
     while( l && *l && !isspace(*l) )
@@ -659,15 +635,16 @@ void doxhdr(char *arg) {
 		arg, l ? l : "" ); /* dubious - if not found, "" */
 	fclose( f );
 	return;
-    }
-
-    if (l && *l) {
+    } else if (l && isdigit(*l) ) {
 	b = a = strtol( l, &l, 10 );
 	if (a && l && *l) {
 	    while (l && isspace(*l))
 		l++;
-	    if ( *l=='-' )
-		b = strtol( ++l, &l, 10 );
+	    if ( *l == '-' ) {
+		l++;
+		if ( l && *l )
+		    b = strtol( l, &l, 10 );
+	    }
 	    while (l && isspace(*l))
 		l++;
 	    if ( l && *l ) {
@@ -686,41 +663,72 @@ void doxhdr(char *arg) {
 	return;
     }
 
+    if ( !arg || !*arg ) {
+	printf( "502 Usage: XHDR header first[-last] or XHDR header message-id\r\n" );
+	return;
+    }
+
     do {
 	n--;
     } while( n>-1 && strncasecmp(arg, h[n], strlen(h[n])) );
 
     markinterest();
 
-    if ( xovergroup != group ) {
-	getxover();
-	xovergroup = group;
-    }
+    if ( a < group->first )
+	a = group->first;
+    else if ( a > group->last )
+	a = group->last;
 
-    if ( a < xfirst )
-	a = xfirst;
-    if ( b > xlast )
-	b = xlast;
-    if ( a > b )
-	a = b;
+    if ( b < a )
+	b = a;
+    else if ( b > group->last )
+	b = group->last;
 
-    printf( "221 %s header (from overview) for postings %lu-%lu:\r\n", 
-	    h[n], a, b );
-    buffer[1023] = '\0';
-    for( c=a; c<=b; c++ ) {
-	if ( xoverinfo[c-xfirst].text ) {
-	    char * l = xoverinfo[c-xfirst].text;
-	    int d;
-	    for( d=0; l && d<=n; d++ )
-		l = strchr(l+1, '\t');
-	    if (l) {
-		char * p;
-		strncpy( buffer, ++l, 1023 );
-		p = strchr(buffer, '\t');
-		if (p)
-		    *p = '\0';
+    if ( n >= 0 ) {
+	printf( "221 %s header (from overview) for postings %lu-%lu:\r\n",
+		h[n], a, b );
+
+	if ( xovergroup != group ) {
+	    getxover();
+	    xovergroup = group;
+	}
+
+	s[1023] = '\0';
+	for( c=a; c<=b; c++ ) {
+	    if ( xoverinfo &&
+		 c >= xfirst && c <= xlast &&
+		 xoverinfo[c-xfirst].text ) {
+		char * l = xoverinfo[c-xfirst].text;
+		int d;
+		for( d=0; l && d<=n; d++ )
+		    l = strchr(l+1, '\t');
+		if (l) {
+		    char * p;
+		    strncpy( s, ++l, 1023 );
+		    p = strchr(s, '\t');
+		    if (p)
+			*p = '\0';
+		}
+		printf("%lu %s\r\n", c, l && *l ? s : "");
+	    }
+	}
+    } else {
+	printf( "221 %s header (from article files) for postings %lu-%lu:\r\n",
+		arg, a, b );
+	n = strlen( arg );
+	for( c=a; c<=b; c++ ) {
+	    sprintf( s, "%lu", c );
+	    f = fopen( s, "r");
+	    while( f ) {
+		l = getaline( f );
+		if ( !l || !*l || *l == '\r' || *l == '\n' ) {
+		    fclose( f );
+		    f = 0;
+		} else if ( l[n] == ':' &&
+			    strncasecmp( arg, l, n ) == 0 ) {
+		    printf("%lu %s\r\n", c, l + n + 2 );
+		}
 	    }
-	    printf("%lu %s\r\n", c, l && *l ? buffer : "");
 	}
     }
 
@@ -787,20 +795,35 @@ void dolistgroup( const char * arg ) {
     struct newsgroup * g;
     unsigned long int art;
 
+    if ( justaftergroup && group && group->name ) {
+	int f;
+	/* mark the previous group as _possibly_ interesting */
+	sprintf( s, "%s/interesting.groups/.%s", spooldir, group->name );
+	if ( (f=open( s, O_WRONLY|O_CREAT|O_TRUNC, 0644)) >= 0 )
+	    close(f);
+    }
+
+    rereadactive();	
+
     if ( arg && strlen(arg) ) {
 	g = findgroup( arg );
 	if ( !g ) {
 	    printf( "481 No such group: %s\r\n", arg );
 	    return;
-	}
-    } else if ( group )
+	} else {
+	    group = g;
+	    artno = g->first;
+	    justaftergroup = TRUE; 
+	}   
+    } else if ( group ) {
 	g = group;
-    else {
+    } else {
 	printf( "481 No group specified\r\n" );
 	return;
     }
 
     markinterest();
+    group = g;
     chdirgroup( g->name );
 
     if ( (xovergroup != group || xovergroup == NULL) && !getxover() ) {
@@ -823,23 +846,10 @@ void dolistgroup( const char * arg ) {
 
 
 int main( int argc, char ** argv ) {
-    char * p;
     int fodder;
     struct hostent *he;
     struct sockaddr_in sa;
 
-    p = strstr( rcsid, ",v " );
-    if ( p ) {
-	strncpy( version, p+3, 9 );
-	p = strchr( version, ' ' );
-	if ( p )
-	    *p = '\0';
-	else
-	    exit( 1 );
-    } else {
-	exit( 1 );
-    }
-
     fodder = sizeof(struct sockaddr_in);
     if (getsockname(0, (struct sockaddr *)&sa, &fodder)) {
 	strcpy( fqdn, "localhost" );
@@ -851,12 +861,14 @@ int main( int argc, char ** argv ) {
 		 he && he->h_name ? he->h_name : inet_ntoa(sa.sin_addr),
 		 63 );
     }
+    if ( strncasecmp( fqdn, "localhost", 9 ) == 0 )
+	whoami();
 
     umask(2);
 
     openlog( "leafnode", LOG_PID|LOG_CONS, LOG_NEWS );
 
-    printf("200 Leafnode NNTP Daemon, RCS version %s running at %s\r\n", 
+    printf("200 Leafnode NNTP Daemon, version %s running at %s\r\n", 
 	   version, fqdn);
     fflush(stdout);
  
@@ -869,13 +881,5 @@ int main( int argc, char ** argv ) {
     if ( nntpout )
 	fprintf( nntpout, "QUIT\r\n" );
 
-    if ( justaftergroup && group && group->name ) {
-	int f;
-	/* mark the previous group as _possibly_ interesting */
-	sprintf( s, "%s/interesting.groups/.%s", spooldir, group->name );
-	if ( (f=open( s, O_WRONLY|O_CREAT|O_TRUNC, 0644)) >= 0 )
-	    close(f);
-    }
-
     exit(0);
 }
