@@ -1,4 +1,4 @@
-/* $Id: texpire.c,v 1.11 1996/06/23 21:33:53 agulbra Exp $
+/* $Id: texpire.c,v 1.14 1997/07/20 00:37:18 agulbra Exp $
 
 Written by Arnt Gulbrandsen <agulbra@troll.no> and copyright 1995
 Troll Tech AS, Postboks 6133 Etterstad, 0602 Oslo, Norway, fax +47
@@ -27,6 +27,15 @@ warranty, or liability of any kind.  */
 
 time_t now;
 
+struct expire_entry {
+    unsigned long xtime;
+    struct expire_entry *next;
+    char group[1];
+};
+
+struct expire_entry *expire_base = NULL;
+unsigned long default_expire;
+
 struct exp {
     char * xover;
     int art;
@@ -34,6 +43,67 @@ struct exp {
 };
 
 
+void free_expire( void ) {
+    
+    struct expire_entry *a, *b;
+
+    b = expire_base;
+    while ((a = b) != NULL) {
+        b = a->next;
+        free(a);
+    }
+}    
+    
+
+void readexpire( void ) {
+    FILE * f;
+    char  *l, *group, *stime, *end;
+    unsigned long i;
+    struct expire_entry *ent, *prev = NULL;
+    
+    sprintf(s, "%s/expireinfo", libdir);
+    if ((f = fopen(s, "r"))) {
+	while ((l = getaline(f))) {
+	    if ((group = strtok(l, " "))) {
+		if ((stime = strtok(NULL, "\n"))) {
+		    if ((i = strtoul(stime, &end, 0))) {
+			i = now-(time_t)(86400*i);
+			ent = (struct expire_entry *)
+			      malloc( sizeof(struct expire_entry) +
+				      strlen( group ) );
+			strcpy(ent->group, group);
+			ent->xtime = i;
+			ent->next = prev;
+			prev = ent;
+		    }
+		}
+	    }
+	    free( l );
+	}
+	fclose(f);
+	expire_base = ent;
+    }
+}    
+
+/*
+05/27/97 - T. Sweeney - Find a group in the expireinfo linked list and return
+                        its expire time. Otherwise, return zero.
+*/
+
+unsigned long lookup_expire(char *group) {
+
+    struct expire_entry *a;
+   
+    a = expire_base;
+    while (a != NULL) {
+        if (!(strcmp(group, a->group)))
+            return a->xtime;
+        a = a->next;
+    }
+    return 0;
+}
+
+
 /* return 1 if xover is a legal overview line, 0 else */
 
 static int legal ( char * xover ) {
@@ -176,7 +246,7 @@ static void dogroup ( struct newsgroup * g ) {
 
     deleted = kept = 0;
     clearidtree();
-
+    
     /* eliminate empty groups */
 
     strcpy( gdir, spooldir );
@@ -285,8 +355,8 @@ static void dogroup ( struct newsgroup * g ) {
 	    /* note that stat() errno is used in the else clause */
 	    if ( (!stat(de->d_name, &st)) && 
 		 (S_ISREG(st.st_mode)) &&
-		 ((st.st_mtime > expire) ||
-		  (st.st_atime > expire)) ) {
+		 (((expire >  0) &&
+		   ((st.st_mtime > expire) || (st.st_atime > expire)))) ) {
 		articles[art-first].kill = 0;
 		p = articles[art-first].xover;
 		for ( n=0; n<4; n++ ) 
@@ -367,7 +437,9 @@ static void dogroup ( struct newsgroup * g ) {
 static void expiregroup( struct newsgroup * g ) {
     if ( g ) {
 	expiregroup( g->right );
-	dogroup( g );
+    if (!(expire = lookup_expire(g->name)))
+        expire = default_expire;
+    dogroup( g );
 	expiregroup( g->left );
     }
 }
@@ -411,8 +483,6 @@ static void expiremsgid( void ) {
 }
 
 
-
-
 int main( int argc, char ** argv ) {
     struct passwd * pw;
 
@@ -431,20 +501,24 @@ int main( int argc, char ** argv ) {
     }
 
     expire = 0;
+
+    readactive();
     readconfig();
+    default_expire = expire;
     if ( expire == 0 ) {
 	fprintf( stderr, "%s: no expire time\n", argv[0] );
 	exit( 2 );
     }
+    free_expire();
 
     openlog( "expire", LOG_PID|LOG_CONS, LOG_NEWS );
 
     now = time(NULL);
 
-    readactive();
     expiregroup( active );
     writeactive();
     expiremsgid();
+    free_expire();
     return 0;
 }
 
