@@ -1,4 +1,4 @@
-/* $Id: util.c,v 1.15 1996/06/23 21:33:53 agulbra Exp $
+/* $Id: util.c,v 1.22 1997/07/23 18:35:18 agulbra Exp $
 
 Written by Arnt Gulbrandsen <agulbra@troll.no> and copyright 1995
 Troll Tech AS, Postboks 6133 Etterstad, 0602 Oslo, Norway, fax +47
@@ -16,12 +16,11 @@ warranty, or liability of any kind. */
 #include <ctype.h>
 #include <errno.h>
 #include <limits.h>
-#include <malloc.h>
+#include <stdlib.h>
 #include <netdb.h>
 #include <setjmp.h>
 #include <signal.h>
 #include <stdio.h>
-#include <stdlib.h>
 #include <string.h>
 #include <syslog.h>
 #include <sys/stat.h>
@@ -32,10 +31,6 @@ warranty, or liability of any kind. */
 
 #include "leafnode.h"
 
-#if !defined(UIO_MAXIOV)
-#define UIO_MAXIOV 8
-#endif
-
 char fqdn[256];
 int verbose = 0;
 
@@ -44,13 +39,68 @@ char s[PATH_MAX+1024]; /* long string, here to cut memory usage */
 FILE *nntpin;
 FILE *nntpout;
 
+unsigned long int nntpport = 0;
+
 static jmp_buf timeout;
 
+char last_command[1025];
+char lineout[1025];
+int authenticated;
+
 static void timer(int sig) {
     longjmp(timeout, 1);
     exit(sig);
 }
 
+
+/*
+05/26/97 - T. Sweeney - Send a string out, keeping a copy in reserve.
+*/
+ 
+void putaline( void ) {
+    strcpy(last_command, lineout);
+    fprintf(nntpout, "%s", lineout);
+    fflush(nntpout);
+}
+
+/* originally from Tim Sweeney
+ *
+ * Returns TRUE if authentication succeeds, FALSE if it does not.
+ *
+ *  Precondition: username != 0.
+ */
+
+int authenticate( void ) {
+    int reply;
+
+    fprintf( nntpout, "authinfo user %s\r\n", username );
+    fflush( nntpout );
+
+    reply = nntpreply();
+    if (reply == 281) {
+	return TRUE;
+    } else if ( reply != 381 ) {
+	syslog( LOG_INFO, "username rejected: %03d", reply);
+	return FALSE;
+    }
+
+    if (password == NULL) {
+	syslog( LOG_INFO, "password needed for authentication" );
+	return FALSE;
+    }
+    fprintf( nntpout, "authinfo pass %s\r\n", password );
+    fflush( nntpout );
+
+    reply = nntpreply();
+
+    if ( reply != 281) {
+	syslog( LOG_INFO, "password failed: %03d", reply);
+	return FALSE;
+    }
+    return TRUE;
+}
+
+
 /*
  * decode an NNTP reply number
  * reads a line from the server and returns an integer
@@ -58,6 +108,8 @@ static void timer(int sig) {
  * 498 is used to mean "protocol error", like smail
  *
  * the text returned is discarded
+ *
+ * from Tim Sweeney: retry in case of authinfo failure.
  */
 
 int nntpreply(void) {
@@ -76,8 +128,8 @@ int nntpreply(void) {
 	    && isdigit(response[1])
 	    && isdigit(response[2])
 	    && ( (response[3]==' ')
-		|| (response[3]=='\0')
-		|| (response[3]=='-') ) ) {
+		 || (response[3]=='\0')
+		 || (response[3]=='-') ) ) {
 	    int rl;
 	    rl = atoi(response);
 	    if (r>0 && r!=rl)
@@ -90,9 +142,19 @@ int nntpreply(void) {
 	    r = 498;	/* protocol error */
 	}
     }
+
+    if (r == 480 && username && !authenticated ) { /* need to authenticate */
+	authenticated = TRUE;
+	if ( authenticate() ) {
+	    fprintf(nntpout, "%s", last_command);
+	    fflush(nntpout);
+	    r = nntpreply();
+	}
+    }
     return r;
 }
 
+
 extern int errno;
 extern int h_errno;
 extern struct state _res;
@@ -102,22 +164,28 @@ extern struct state _res;
 int nntpconnect( void ) {
     struct hostent *hp;
     struct servent *sp;
+    struct servent sp_def;
     struct sockaddr_in s_in;
     int sock;
     register int i;
 
     memset((void *)&s_in, 0, sizeof(s_in));
-    sp = getservbyname("nntp", "tcp");
-    if (sp == NULL) {
-	syslog( LOG_ERR, "unable to find service NNTP" );
-	return FALSE;
+    if (nntpport == 0) {
+	sp = getservbyname("nntp", "tcp");
+	if (sp == NULL) {
+	    syslog( LOG_ERR, "unable to find service NNTP" );
+	    return FALSE;
+	}
+    } else {
+	sp=&sp_def;
+	sp->s_port=htons(nntpport);
     }
 
-/* Fetch the ip addresses of the given host. */
+    /* Fetch the ip addresses of the given host. */
     hp = gethostbyname( upstream );
     if (hp) {
 
-/* Try to make connection to each of the addresses in turn. */
+	/* Try to make connection to each of the addresses in turn. */
 	for (i = 0; (int *)(hp->h_addr_list)[i]; i++) {
 	    s_in.sin_family = hp->h_addrtype;
 	    s_in.sin_port = sp->s_port;
@@ -147,11 +215,11 @@ int nntpconnect( void ) {
 		break;
 
 	    switch(nntpreply()) {
-		case 200: 
-		case 201: 
-		    syslog( LOG_INFO, "connected to %s",
-			    inet_ntoa( s_in.sin_addr ) );
-		    return TRUE;
+	    case 200: 
+	    case 201: 
+		syslog( LOG_INFO, "connected to %s",
+			inet_ntoa( s_in.sin_addr ) );
+		return TRUE;
 	    }
 	    shutdown(fileno(nntpout), 0);
 	}/* end of IP-addresses for loop */
@@ -387,7 +455,14 @@ void insertgroup(const char * name,
 		    (*a)->last = last;
 		if (server >= 0)
 		    (*a)->server = server;
-		(*a)->desc = (desc && *desc) ? strdup( desc ) : "";
+		if (desc && *desc) {
+		    /* This leaks memory if the new desc > old one, but I 
+		       cannot see a way round this... */
+		    if (strlen((*a)->desc) >= strlen(desc))
+			strcpy((*a)->desc, desc);
+		    else
+			(*a)->desc = strdup( desc );
+		}
 		return;
 	    }
 	} else {
@@ -507,16 +582,15 @@ void readactive( void ) {
     p = stuff;
     while ( p && *p ) {
 	q = p;
-	while (p && *p && !isspace(*p))
-	    p++;
-	if (*p)
+	p = strchr(p, ' ');
+	if (p && *p)
 	    *p++ = '\0';
 	if ( *q ) {
 	    last = p ? strtol( p, &p, 10 ) : 0;
 	    first = p ? strtol( p, &p, 10 ) : 0;
 	    server = p ? strtol( p, &p, 10 ) : 0;
-	    while (p && *p && isspace(*p) && p++)
-		;
+	    while (p && *p && isspace(*p))
+		p++;
 	    if (first < 1)
 		first = 1;
 
@@ -707,17 +781,46 @@ void clearidtree( void ) {
 }
 
 
+struct newsgroup * getexpire( struct newsgroup * g, char * l) {
+    struct newsgroup * dummy = NULL;
+
+    if (g) {
+        if (strcasecmp( l, g->name) == 0)
+            dummy = g;
+	else if ( (dummy = getexpire(g->left, l)) == NULL )
+	    dummy = getexpire(g->right, l);
+    }
+    return dummy;
+}
+
+
+/* 
+05/25/97 - T. Sweeney - Modified to read user name and password for AUTHINFO.
+			Security questionable as password is stored in
+			plaintext in insecure file.
+*/
+
+/* this thing -really- needs a redo.  that many parameters means that
+   cut and paste is not a good strategy any more */
 const char * upstream;
+const char * username;
+const char * password;
 int artlimit = 0;
+int initiallimit = 0;
 time_t expire;
 struct supplement * supplement;
+struct newsgroup * g;
 
 void readconfig( void ) {
+    struct supplement * ns;
     FILE * f;
     char * l;
     char * p;
 
+    ns = supplement;
     upstream = NULL;
+    username = NULL;
+    password = NULL;
     artlimit = 0;
     
     sprintf( s, "%s/config", libdir );
@@ -748,7 +851,44 @@ void readconfig( void ) {
 			    "unable to parse %s (server line)", s );
 		    exit( 2 );
 		}
-
+	    } else if ( strncasecmp( l, "username", 8 )==0 ) {
+		l += 8;
+		while ( isspace(*l) )
+		    l++;
+		if ( *l=='=' ) {
+		    l++;
+		    while ( isspace(*l) )
+			l++;
+		    p = l;
+		    while ( p && *p && !isspace(*p) )
+			p++;
+		    if ( p )
+			*p = '\0';
+		    username = strdup( l );
+		} else {
+		    syslog( LOG_ERR, 
+			    "unable to parse %s (username line)", s );
+		    exit( 2 );
+		}
+	    } else if ( strncasecmp( l, "password", 8 )==0 ) {
+		l += 8;
+		while ( isspace(*l) )
+		    l++;
+		if ( *l=='=' ) {
+		    l++;
+		    while ( isspace(*l) )
+			l++;
+		    p = l;
+		    while ( p && *p && !isspace(*p) )
+			p++;
+		    if ( p )
+			*p = '\0';
+		    password = strdup( l );
+		} else {
+		    syslog( LOG_ERR, 
+			    "unable to parse %s (password line)", s );
+		    exit( 2 );
+		}
 	    } else if ( strncasecmp( l, "expire", 6 )==0 ) {
 		l += 6;
 		while ( isspace(*l) )
@@ -803,6 +943,60 @@ void readconfig( void ) {
 			    "unable to parse %s (maxfetch has no =)", s );
 		    exit( 2 );
 		}
+	    } else if ( strncasecmp( l, "port", 4 )==0 ) {
+		l += 4;
+		while ( isspace(*l) )
+		    l++;
+		if ( *l=='=' ) {
+		    unsigned long i;
+		    l++;
+		    while ( isspace(*l) )
+			l++;
+		    p = l;
+		    while ( p && *p && !isspace(*p) )
+			p++;
+		    if ( p )
+			*p = '\0';
+		    i = strtol( l, &p, 0 );
+		    if ( l && *l && p && !*p )
+			nntpport = i;
+		    else {
+			syslog( LOG_ERR,
+				"unable to parse %s (port number)", s );
+			exit( 2 );
+		    }
+		} else {
+		    syslog( LOG_ERR,
+			    "unable to parse %s (port has no =)", s );
+		    exit( 2 );
+		}
+	    } else if ( strncasecmp( l, "initialfetch", 12 )==0 ) {
+		l += 12;
+		while ( isspace(*l) )
+		    l++;
+		if ( *l=='=' ) {
+		    unsigned long i;
+		    l++;
+		    while ( isspace(*l) )
+			l++;
+		    p = l;
+		    while ( p && *p && !isspace(*p) )
+			p++;
+		    if ( p )
+			*p = '\0';
+		    i = strtoul( l, &p, 0 );
+		    if ( l && *l && p && !*p )
+			initiallimit = i;
+		    else {
+			syslog( LOG_ERR,
+				"unable to parse %s (initialfetch number)", s );
+			exit( 2 );
+		    }
+		} else {
+		    syslog( LOG_ERR, 
+			    "unable to parse %s (initialfetch has no =)", s );
+		    exit( 2 );
+		}
 	    } else if ( strncasecmp( l, "supplement", 10 )==0 ) {
 		l += 10;
 		while( isspace(*l) || *l=='=' )
@@ -812,14 +1006,19 @@ void readconfig( void ) {
 		    while ( isalnum( *l ) || *l=='.' || *l=='-' || *l=='/' )
 			l++;
 		    if ( b != l ) {
-			struct supplement * ns = (struct supplement *) critmalloc( sizeof(struct supplement) + (int)l-(int)b, "allocating space for supplementary server name" );
+			struct supplement * ns
+			    = (struct supplement *)
+			    critmalloc( sizeof(struct supplement) +
+					(int)l-(int)b,
+					"allocating space for supplementary"
+					" server name" );
 			*l = '\0';
 			ns->next = supplement;
 			strcpy( ns->name, b );
 			supplement = ns;
 		    } else {
 			syslog( LOG_ERR, 
-			    "impossibility 122", s );
+				"impossibility 122", s );
 			exit( 2 );
 		    }
 		} else {
@@ -1033,7 +1232,7 @@ void stripspace( char * p ) {
     char * p2;
 
     p1 = p2 = p;
-    while (p1 && *p1) {
+    while ( *p1 ) {
 	if ( isspace(*p1) ) {
 	    *p2 = ' ';
 	    do {
@@ -1134,6 +1333,8 @@ char * getxoverline( const char * filename ) {
 		    xref = strdup( l );
 		    h = &xref;
 		}
+	    } else {
+		h = NULL;
 	    }
 	} while ( l && !feof(f) );
 	if ( from && date && subject && msgid && bytes && lines && body ) {
@@ -1217,7 +1418,6 @@ int getxover( void ) {
 	return 0;
     }
 
-
     xfirst = INT_MAX;
     xlast = 0;
     while ( (de=readdir(d)) ) {
@@ -1229,6 +1429,7 @@ int getxover( void ) {
 		xlast = art;
 	}
     }
+
     if ( xlast < xfirst ) {
 	closedir( d );
 	return 0;
@@ -1392,6 +1593,93 @@ int getxover( void ) {
 }
 
 
+
+/* build and return an open fd to pseudoart in group */
+FILE * buildpseudoart(const char * grp) {
+    FILE * f;
+
+    f = tmpfile();
+    if (!f)
+	return f;
+
+    fprintf(f, "Path: %s\n", fqdn);
+    fprintf(f, "Newsgroups: %s\n", grp);
+    fprintf(f, "From: Leafnode <nobody@%s>\n", fqdn);
+    fprintf(f, "Subject: Leafnode placeholder for group %s\n", grp);
+    fprintf(f, "Date: %s\n", rfctime() );
+    fprintf(f, "Message-ID: <leafnode:placeholder:%s@%s>\n", grp, fqdn);
+    fprintf(f, "\n");
+    fprintf(f, "This server is running leafnode, which is a dynamic NNTP proxy\n");
+    fprintf(f, "This means that it does not retrieve newsgroups unless someone is\n");
+    fprintf(f, "actively reading them.\n");
+    fprintf(f, "\n");
+    fprintf(f, "If you do an operation on a group - such as reading an article,\n");
+    fprintf(f, "looking at the group table of contents or similar, then leafnode\n");
+    fprintf(f, "will go and fetch articles from that group when it next updates.\n");
+    fprintf(f, "\n");
+    fprintf(f, "If you see articles in groups you do not read, that is almost\n");
+    fprintf(f, "always because of cross-posting.  These articles do not occupy any\n");
+    fprintf(f, "more space - they are hard-linked into each newsgroup directory\n");
+    fprintf(f, "\n");
+
+    fprintf(f, "Since you have read this dummy article, leafnode will retrieve this");
+    fprintf(f, "group on the next update.  Please look in this group a little later\n");
+    fprintf(f, "and you should see some articles.\n");
+    fprintf(f, "\n");
+    fprintf(f, "If you have any queries about this, please talk to your newsmaster.\n");
+    fprintf(f, "\n");
+    fprintf(f, "Leafnode can be found at\n");
+    fprintf(f, "\thttp://www.troll.no/freebies/leafnode.html\n");
+    fprintf(f, "\n");
+    fprintf(f, "\n");
+
+    rewind(f);
+    return f;
+}
+
+
+const char* rfctime(void) {
+    static char date[128];
+    const char * months[] = { "Jan", "Feb", "Mar", "Apr",
+	"May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
+    const char * days[] = { "Sun", "Mon", "Tue", "Wed",
+	"Thu", "Fri", "Sat" };
+    time_t now;
+    struct tm gmt;
+    struct tm local;
+    struct tm conv;
+    signed int diff;
+
+    /* get local and Greenwich times */
+    now = time(0);
+    gmt	  = *(gmtime(&now));
+    local = *(localtime(&now));
+    conv  = *(localtime(&now));
+
+    /* if GMT were local, what would it be? */
+    conv.tm_year = gmt.tm_year;
+    conv.tm_mon	 = gmt.tm_mon;
+    conv.tm_mday = gmt.tm_mday;
+    conv.tm_hour = gmt.tm_hour;
+    conv.tm_min	 = gmt.tm_min;
+    conv.tm_sec	 = gmt.tm_sec;
+    conv.tm_wday = gmt.tm_wday;
+    conv.tm_yday = gmt.tm_yday;
+
+    /* see how far the real local is from GMT-as-local */
+    /* should convert to signed int from double properly */
+    diff = (mktime(&local) - mktime(&conv)) / 60.0;
+
+    /* finally print the string */
+    sprintf(date, "%3s, %d %3s %4d %02d:%02d:%02d %+03d%02d",
+	    days[local.tm_wday], local.tm_mday, months[local.tm_mon],
+	    local.tm_year+1900, local.tm_hour, local.tm_min, local.tm_sec, 
+	    diff/60, diff%60);
+
+    return (date);
+}
+
+
 #ifndef HAVE_STRDUP
 char * strdup(const char *s) {
     char * s1;
